(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2017)
License: LGPL (see file COPYING.LESSER)
*)
theory ACU
  imports AC_Rewriting_Base
begin

locale abstract_unf =
  fixes nf :: "('f, 'v) term \<Rightarrow> 'a"
    and F :: "'f set"
    and U :: "'f \<Rightarrow> 'f set"
  assumes nf_unit_right [simp]: "f \<in> F \<Longrightarrow> u \<in> U f \<Longrightarrow> nf (Bin f t (Fun u [])) = nf t"
    and nf_FunI [intro]: "map nf ss = map nf ts \<Longrightarrow> nf (Fun f ss) = nf (Fun f ts)"
begin

lemma nf_BinI:
  "nf s = nf s' \<Longrightarrow> nf t = nf t' \<Longrightarrow> nf (Bin f s t) = nf (Bin f s' t')"
  by (auto)

lemma nf_eq_ctxt:
  "nf s = nf t \<Longrightarrow> nf (C\<langle>s\<rangle>) = nf (C\<langle>t\<rangle>)"
  by (induct C) auto

end

abbreviation "U_rule f t u \<equiv> (Bin f t (Fun u []), t)"
definition "U_rules F U = {U_rule f t u | f t u. f \<in> F \<and> u \<in> U f}"
abbreviation "ustep F U \<equiv> rstep (U_rules F U)"
definition "U_class F U s = {t. (s, t) \<in> (ustep F U)\<^sup>\<leftrightarrow>\<^sup>*}"

lemma ctxt_closed_ustep_symcl: "ctxt.closed ((ustep F U)\<^sup>\<leftrightarrow>)" by auto

lemmas
  args_uconv_imp_uconv = args_steps_imp_steps [OF ctxt_closed_ustep_symcl, folded conversion_def]

lemma U_rule_in_U_rules [intro]:
  "f \<in> F \<Longrightarrow> u \<in> U f \<Longrightarrow> U_rule f t u \<in> U_rules F U"
  by (auto simp: U_rules_def)

lemma U_class_eq_conv [simp]:
  "U_class F U s = U_class F U t \<longleftrightarrow> (s, t) \<in> (ustep F U)\<^sup>\<leftrightarrow>\<^sup>*"
by (auto simp: U_class_def) (metis conversion_def conversion_inv rtrancl_trans)+

interpretation U_class_abstract_unf: abstract_unf where nf = "U_class F U"
  by (unfold_locales) (auto simp: map_eq_conv' args_uconv_imp_uconv)

context
  fixes F :: "'f set" and U :: "'f \<Rightarrow> 'f set"
begin

fun uBin :: "'f \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "uBin f s (Fun u []) = (if u \<in> U f then s else Bin f s (Fun u []))"
| "uBin f s t = Bin f s t"

function unf :: "('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "unf (Var x) = Var x"
| "unf (Bin f s t) = (if f \<in> F then uBin f (unf s) (unf t) else Bin f (unf s) (unf t))"
| "\<And>f ts. \<forall>u v. ts \<noteq> [u, v] \<Longrightarrow> unf (Fun f ts) = Fun f (map unf ts)"
by (auto, atomize_elim, insert Bin_cases, auto)
termination by (lexicographic_order)

function is_unf :: "('f, 'v) term \<Rightarrow> bool"
where
  "is_unf (Var x) \<longleftrightarrow> True"
| "is_unf (Bin f s t) \<longleftrightarrow> is_unf s \<and> is_unf t \<and> (f \<in> F \<longrightarrow> (\<forall>u \<in> U f. t \<noteq> Fun u []))"
| "\<And>f ts. \<forall>u v. ts \<noteq> [u, v] \<Longrightarrow> is_unf (Fun f ts) \<longleftrightarrow> (\<forall>t \<in> set ts. is_unf t)"
by (auto, atomize_elim, insert Bin_cases, auto)
termination by (lexicographic_order)

lemma unf_code [code]:
  "unf (Var x) = Var x"
  "unf (Bin f s t) = (if f \<in> F then uBin f (unf s) (unf t) else Bin f (unf s) (unf t))"
  "unf (Fun f []) = Fun f []"
  "unf (Fun f [t]) = Fun f [unf t]"
  "unf (Fun f (s # t # u # us)) = Fun f (unf s # unf t # unf u # map unf us)"
by simp_all

end

lemma uBin_unit_right [simp]:
  "uBin U f s t = (if \<exists>u \<in> U f. t = Fun u [] then s else Bin f s t)"
  by (induct f s t rule: uBin.induct) simp_all

lemma is_unf_uBin [intro]:
  "is_unf F U s \<Longrightarrow> is_unf F U t \<Longrightarrow> is_unf F U (uBin U f s t)"
  by (induct f s t rule: uBin.induct) auto

lemma is_unf_unf [intro]: "is_unf F U (unf F U t)"
proof (induct t rule: unf.induct [of _ F])
  case (3 f ts)
  then show ?case using Bin_cases [of "Fun f (map (unf F U) ts)"] by auto
qed auto

interpretation unf_abstract_unf: abstract_unf where nf = "unf F U"
proof (unfold_locales)
  fix f and ss ts :: "('a, 'b) term  list"
  assume "map (unf F U) ss = map (unf F U) ts"
  then show "unf F U (Fun f ss) = unf F U (Fun f ts)"
    using Bin_cases [of "Fun f ss"] and Bin_cases [of "Fun f ts"] by force
qed auto

context abstract_unf
begin

lemma ustep_imp_nf_eq:
  "(s, t) \<in> ustep F U \<Longrightarrow> nf s = nf t"
  by (force intro: nf_eq_ctxt simp: U_rules_def)

lemma uconv_imp_nf_eq:
  "(s, t) \<in> (ustep F U)\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow> nf s = nf t"
  unfolding conversion_def
  by (induct rule: rtrancl_induct) (auto simp: ustep_imp_nf_eq )

lemma nf_uBin [simp]:
  "f \<in> F \<Longrightarrow> nf (uBin U f s t) = nf (Bin f s t)"
  by (induct f s t rule: uBin.induct) auto

lemma nf_unf: "nf (unf F U t) = nf t"
  apply (induct t rule: unf.induct [of _ F])
    apply auto
   apply (metis nf_BinI nf_unit_right)
  by force

end

lemma uconv_iff [code_unfold]:
  "(s, t) \<in> (ustep F U)\<^sup>\<leftrightarrow>\<^sup>* \<longleftrightarrow> unf F U s = unf F U t"
proof
  assume "unf F U s = unf F U t"
  then have "U_class F U (unf F U s) = U_class F U (unf F U t)" by simp
  then show "(s, t) \<in> (ustep F U)\<^sup>\<leftrightarrow>\<^sup>*"
    by (simp add: U_class_abstract_unf.nf_unf)
qed (simp add: unf_abstract_unf.uconv_imp_nf_eq)

end
