(*
Author:  Sebastiaan Joosten (2016-2017)
Author:  René Thiemann (2016-2017)
Author:  Akihisa Yamada (2016-2017)
License: LGPL (see file COPYING.LESSER)
*)
theory Sorted_Formula
imports
  "Abstract-Rewriting.Abstract_Rewriting"
  TRS.Trs_Impl
  First_Order_Terms.Option_Monad
  Sorted_Algebra
  Formula
begin

type_synonym ('v,'t,'d) valuation = "'v \<times> 't \<Rightarrow> 'd"


abbreviation rename_vars_formula :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v,'t) exp formula \<Rightarrow> ('f,'w,'t) exp formula"
where "rename_vars_formula r cls \<equiv> map_formula (rename_vars r) cls"

adhoc_overloading rename_vars rename_vars_formula

locale prelogic =
  pre_sorted_algebra where type_fixer = type_fixer
  for type_fixer :: "('f\<times>'t\<times>'d) itself"
  and Bool_types :: "'t set"
  and to_bool :: "'d \<Rightarrow> bool"
begin

definition "literal e \<equiv> (\<exists>b \<in> Bool_types. e :\<^sub>f b) \<and> \<not> is_Var e"

lemma literal_simps[simp]: "literal (Var x) = False"
  "literal (Fun f es) = (\<exists>b \<in> Bool_types. Fun f es :\<^sub>f b)" by (auto simp: literal_def)

abbreviation "BOOL_EXP \<equiv> Collect literal"

abbreviation "eval_lit \<alpha> t \<equiv> to_bool (\<lbrakk>t\<rbrakk>\<alpha>)"

sublocale formula literal assignment eval_lit.

lemma literal_imp_expression[simp]: "literal e \<Longrightarrow> expression e" unfolding literal_def by auto

lemma formula_rename_vars[simp]: "formula (rename_vars r e) = formula e"
proof (induct e)
  case (Literal x)
  then show ?case by (cases x, auto)
qed auto

lemma formula_rename_vars_map[simp]:
  "list_all formula (map (rename_vars r) l) = list_all formula l"
  using formula_rename_vars unfolding list_all_def by auto

lemma eval_with_fresh_var[simp]: "xx \<notin> vars_term e \<Longrightarrow> \<lbrakk>e\<rbrakk> (\<alpha>(xx:=a)) = \<lbrakk>e\<rbrakk> \<alpha>" by auto

lemma satisfies_with_fresh_var[simp]:
  assumes fv: "xx \<notin> vars_formula \<phi>"
  shows "\<alpha>(xx:=a) \<Turnstile> \<phi> \<longleftrightarrow> \<alpha> \<Turnstile> \<phi>" by (insert fv, induct \<phi>, auto)

lemma satisfies_rename_vars:
  assumes r: "\<And> x t. \<alpha> (x,t) = \<beta> (r x, t)"
  shows "\<beta> \<Turnstile> (rename_vars r \<phi>) \<longleftrightarrow> \<alpha> \<Turnstile> \<phi>"
  by (induct \<phi>, auto simp: eval_rename_vars[of \<alpha>, OF r])

end

locale logic = prelogic + sorted_algebra


subsection \<open>Checking validity\<close>

type_synonym ('h,'f,'v,'t) logic_checker =
  "'h \<Rightarrow> ('f,'v,'t) exp formula \<Rightarrow> showsl check"

datatype 'h hint =
  Default
| Base 'h
| Distribute nat "('h hint list)"
| Erase nat "('h hint)"
| LexStrict "('h hint list)"
| LexWeak "('h hint list)"

primrec showsl_hint where
  "showsl_hint Default = showsl_lit (STR '' auto'')"
| "showsl_hint (Base raw) = showsl_lit (STR '' '') o showsl raw"
| "showsl_hint (Distribute i hints) = showsl_lit (STR '' Distribute '') o showsl i o 
    showsl_list_gen id (STR '' []'') (STR '' ['') (STR '', '') (STR '']'') (map showsl_hint hints)"
| "showsl_hint (LexStrict hints) = showsl_lit (STR '' LexStrict'') o 
    showsl_list_gen id (STR '' []'') (STR '' ['') (STR '', '') (STR '']'') (map showsl_hint hints)"
| "showsl_hint (LexWeak hints) = showsl_lit (STR '' LexWeak'') o 
    showsl_list_gen id (STR '' []'') (STR '' ['') (STR '', '') (STR '']'') (map showsl_hint hints)"
| "showsl_hint (Erase i hint) = showsl_lit (STR '' Erase '') o showsl i o showsl_lit (STR '' ['') o showsl_hint hint o showsl_lit (STR '' ]'')"


instantiation hint :: (default) default
begin
  definition "default \<equiv> Default"
  instance ..
end

hide_const(open) Default Base Distribute Erase

locale pre_trivial_checker =
  fixes trivial_lit_or :: "'l \<Rightarrow> 'l \<Rightarrow> bool"
begin

fun trivial_checker_lit where
  "trivial_checker_lit [] = False"
| "trivial_checker_lit (l#ls) \<longleftrightarrow> (\<exists> n \<in> set ls. trivial_lit_or l n) \<or> trivial_checker_lit ls"

definition trivial_clause_checker
  where "trivial_clause_checker f = (case f of Disjunction fs \<Rightarrow>
    trivial_checker_lit (map get_Literal (filter is_Literal fs)))" 

lemma trivial_checker_lit:
  assumes "trivial_checker_lit lits"
  shows "\<exists> a b. a \<in> set lits \<and> b \<in> set lits \<and> trivial_lit_or a b"
  by (insert assms, induct lits, auto)

end

locale trivial_checker =
  formula where literal = "literal :: 'l \<Rightarrow> bool" +
  pre_trivial_checker where trivial_lit_or = "trivial_lit_or::'l \<Rightarrow> 'l \<Rightarrow> bool"
  for trivial_lit_or literal +
  assumes trivial_lit_or: "trivial_lit_or a b \<Longrightarrow> eval_lit \<alpha> a \<or> eval_lit \<alpha> b"
begin

lemma trivial_clause_checker:
  assumes "is_disj_lits \<phi>" and "trivial_clause_checker \<phi>" shows "\<Turnstile> \<phi>"
  using assms
proof (intro validI, induct \<phi>)
  case (1 \<phi>s)
  from 1(3) obtain x and y where "Literal x \<in> set \<phi>s" and "Literal y \<in> set \<phi>s"
    and "trivial_lit_or x y"
    by (auto simp: trivial_clause_checker_def elim!: is_Literal.elims dest!: trivial_checker_lit)
  with 1(2) show ?case
    by (cases "eval_lit \<alpha> x"; cases "eval_lit \<alpha> y", auto dest: trivial_lit_or)
qed

end

locale pre_logic_checker =
  prelogic where type_fixer = "TYPE('f\<times>'t\<times>'d)" +
  pre_trivial_checker where trivial_lit_or = "trivial_lit_or::('f,'v :: showl,'t) exp \<Rightarrow> _" +
  showsl_formula "showsl_atom :: ('f,'v,'t) exp \<Rightarrow> showsl"
  for showsl_atom trivial_lit_or +
  fixes type_fixer :: "('f \<times> 'v \<times> 't \<times> 'd \<times> 'h :: {default,showl}) itself"
    and check_clause :: "('h,'f,'v,'t) logic_checker"
begin

definition "check_valid_formula \<phi> \<equiv>
  (case flatten \<phi> of Conjunction \<phi>s \<Rightarrow> check_allm (
     \<lambda> \<phi>. try (check (trivial_clause_checker \<phi>) (STR ''trivial clause checker failed'')) catch (\<lambda> _. 
       case flatten ( \<phi>) of Conjunction \<phi>s \<Rightarrow> check_allm (check_clause default) \<phi>s)) \<phi>s)
  <+? (\<lambda>s. showsl_lit (STR ''problem in checking validity of formula '') o showsl_formula \<phi> o showsl_nl o s)"

fun is_or_and_shape where
  "is_or_and_shape (Disjunction [\<phi>1,Conjunction [\<phi>2,\<phi>3]]) = True"
| "is_or_and_shape _ = False"

fun is_conj_shape where
  "is_conj_shape [hint1,hint2] (Conjunction [\<phi>1,\<phi>2]) = True"
| "is_conj_shape _ _ = False"

fun is_disj_shape where
  "is_disj_shape (hint1#hints) (Disjunction [Conjunction [\<phi>1,\<phi>2], Conjunction [\<phi>3,\<phi>4]]) = True"
| "is_disj_shape _ _ = False"

function(sequential) check_formula and check_formula_dist and check_formula_lex_weak and check_formula_lex
where
  "check_formula hint.Default \<phi> = check_valid_formula \<phi>"
| "check_formula (hint.Base h) \<phi> = do {
    let \<psi> = (simplify \<phi>);
    check (is_disj_lits \<psi>) (showsl (STR ''base hint given to '') o showsl_formula \<psi>);
    check_clause h \<psi> <+? (\<lambda> s. showsl (STR ''problem in checking '') o showsl_formula \<phi> o showsl_nl o s)}"
| "check_formula (hint.Erase n hint) \<phi> =
    ( case \<phi> of formula.Disjunction \<phi>s \<Rightarrow> do {
        let l = length \<phi>s;
        check (n < l) (showsl (STR ''erase hint at position '') o showsl n o showsl (STR '' while goal is length '') o showsl l);
        let pre = take n \<phi>s;
        let post = drop (Suc n) \<phi>s;
        check_formula hint (formula.Disjunction (pre @ False\<^sub>f # post))
      }
      | _ \<Rightarrow> error (showsl (STR ''Erase hint to non-Disjunction''))
    )"
| "check_formula (hint.Distribute n hints) \<phi> =
    ( case \<phi> of formula.Disjunction \<phi>s \<Rightarrow> do {
        let l = length \<phi>s;
        check (n < l) (showsl (STR ''distribute hint at position '') o showsl n o showsl (STR '' while goal is length '') o showsl l);
        let pre = take n \<phi>s;
        let post = drop (Suc n) \<phi>s;
        case \<phi>s ! n of
          Conjunction \<psi>s \<Rightarrow> check_formula_dist hints pre post \<psi>s
        | _ \<Rightarrow> error (showsl (STR ''Distribute hint in: \<newline>'') o showsl_formula \<phi> o showsl (STR ''\<newline> at non-Conjunction position:\<newline>'') o showsl_formula (\<phi>s ! n))
      }
      | _ \<Rightarrow> error (showsl (STR ''Distribute hint for non Disjunction '') o showsl_formula \<phi>)
    )"
| "check_formula (hint.LexWeak hints) \<psi> = (case \<psi> of 
     Disjunction (\<phi>#\<phi>s) \<Rightarrow> check_formula_lex_weak hints \<phi> \<phi>s
   | _ \<Rightarrow> error (showsl (STR ''LexWeak hint applied on '') o showsl_formula \<psi>))"
| "check_formula (hint.LexStrict hints) \<psi> = (case \<psi> of 
     Disjunction (\<phi>#\<phi>s) \<Rightarrow> check_formula_lex hints \<phi> \<phi>s
   | _ \<Rightarrow> error (showsl (STR ''LexStrict hint applied on '') o showsl_formula \<psi>))"
| "check_formula_dist [] pre post [] = succeed"
| "check_formula_dist (h#hs) pre post (\<phi>#\<phi>s) =
   check_formula h (formula.Disjunction (pre @ \<phi> # post)) \<then> check_formula_dist hs pre post \<phi>s"
| "check_formula_dist _ _ _ _ = error (showsl (STR ''Length mismatch in Distribute hints''))"
| "check_formula_lex_weak [] \<phi> \<phi>s =
   check (trivial_formula \<phi>) (showsl (STR ''LexWeak base case error: '') o showsl_formula \<phi>)"
| "check_formula_lex_weak (hint#hints) \<phi> \<phi>s = (if is_or_and_shape \<phi> then (case \<phi> of
    Disjunction [\<phi>1,Conjunction [\<phi>2,\<phi>3]] \<Rightarrow> do {
      check_formula hint (Disjunction (\<phi>2#\<phi>s));
      check_formula_lex_weak hints \<phi>3 \<phi>s
    })
    else (case hints of 
      Nil \<Rightarrow> check_formula hint (Disjunction (\<phi>#\<phi>s))
    | _ \<Rightarrow> error (showsl (STR ''LexWeak hint application error''))))"
| "check_formula_lex hints \<phi> \<phi>s = (if is_conj_shape hints \<phi> then
      (case hints of [hint1,hint2] \<Rightarrow> case \<phi> of Conjunction [\<phi>1,\<phi>2] \<Rightarrow> do {
      check_formula hint1 (Disjunction (\<phi>1#\<phi>s));
      check_formula hint2 (Disjunction (\<phi>2#\<phi>s))
    }) else (if is_disj_shape hints \<phi> then
      (case hints of hint1#hints2 \<Rightarrow> case \<phi> of Disjunction [Conjunction [\<phi>1,\<phi>2], Conjunction [\<phi>3,\<phi>4]] \<Rightarrow> 
        (case hints2 of 
          [hint2] \<Rightarrow> do {
            check_formula hint1 (Disjunction (\<phi>1#\<phi>s));
            check_formula hint2 (Disjunction (\<phi>2#\<phi>s))
          }
        | _ \<Rightarrow> do {
            check_formula hint1 (Disjunction (\<phi>3#\<phi>s));
            check_formula_lex hints2 \<phi>4 \<phi>s
          })) 
      else error (showsl (STR ''LexStrict hint application error: '') o
        showsl_hint (LexStrict hints) o showsl (STR ''\<newline>applied on '') o showsl_formula \<phi>)))" 
  by (pat_completeness,auto)

termination by (relation "measure (\<lambda>x.
  case x of Inl (Inl (x,_)) \<Rightarrow> size x
  | Inl (Inr (xs,_)) \<Rightarrow> size_list size xs
  | Inr (Inl (xs,_)) \<Rightarrow> size_list size xs
  | Inr (Inr (xs,_)) \<Rightarrow> size_list size xs)", auto)

lemma isOK_check_formula_dist:
  assumes "isOK (check_formula_dist hs pre post \<phi>s)" and "\<phi> \<in> set \<phi>s"
  shows "\<exists>h \<in> set hs. isOK (check_formula h (Disjunction (pre @ \<phi> # post)))"
proof (insert assms, induct \<phi>s arbitrary: hs)
  case IH: (Cons a \<phi>s) then show ?case by(cases hs, auto)
qed auto

end

locale logic_checker =
  logic where type_fixer = "TYPE('f\<times>'t\<times>'d)" +
  pre_logic_checker
    where trivial_lit_or = "trivial_lit_or::('f,'v::showl,'t) exp \<Rightarrow> _"
      and check_clause = "check_clause :: ('h::{default,showl},'f,'v,'t) logic_checker"
 +
  trivial_checker
    where trivial_lit_or = trivial_lit_or
      and literal=literal
      and assignment=assignment
      and eval_lit="\<lambda>\<alpha> t. to_bool (\<lbrakk>t\<rbrakk>\<alpha>)"
  for trivial_lit_or check_clause
 +
  assumes check_clause:
    "\<And>h \<phi>. isOK(check_clause h \<phi>) \<Longrightarrow> formula \<phi> \<Longrightarrow> is_disj_lits \<phi> \<Longrightarrow> \<Turnstile> \<phi>"
begin

lemma check_valid_formula:
  assumes ok: "isOK(check_valid_formula \<phi>)" and \<phi>: "formula \<phi>"
  shows "\<Turnstile> \<phi>"
proof-
  from \<phi> have form: "formula (flatten \<phi>)" by auto
  from is_cnf_flatten[of \<phi>] have "is_cnf (flatten \<phi>)" .
  then obtain \<phi>s where *: "flatten \<phi> = Conjunction \<phi>s" "\<And> \<phi>\<^sub>i. \<phi>\<^sub>i \<in> set \<phi>s \<Longrightarrow> is_disj_lits \<phi>\<^sub>i" by (cases, auto)
  from form have form: "\<And>\<psi>. \<psi> \<in> set \<phi>s \<Longrightarrow> formula \<psi>" by (auto simp: *)
  {
    fix \<psi>
    assume **: "\<psi> \<in> set \<phi>s" 
    note form = form[OF **]
    note clause = *(2)[OF **]
    from is_cnf_flatten[of \<psi>] have "is_cnf (flatten \<psi>)" .
    then obtain \<chi>s where 1: "flatten \<psi> = Conjunction \<chi>s" "\<And> \<phi>\<^sub>i. \<phi>\<^sub>i \<in> set \<chi>s \<Longrightarrow> is_disj_lits \<phi>\<^sub>i" by (cases, auto)
    from form have form': "\<And>\<psi>. \<psi> \<in> set \<chi>s \<Longrightarrow> formula \<psi>" 
      using 1 formula_flatten by force
    from ok[unfolded check_valid_formula_def *, simplified, rule_format, OF **, unfolded 1, simplified]
    consider (Triv) "trivial_clause_checker \<psi>" | (lc) "\<forall>x\<in>set \<chi>s. isOK (check_clause default x)" by auto
    then have "\<Turnstile> \<psi>" 
    proof (cases)
      case lc
      {
        fix \<chi>
        assume chi: "\<chi> \<in> set \<chi>s" 
        with lc have ok: "isOK(check_clause default \<chi>)" by auto
        from 1(2)[OF chi] have "is_disj_lits \<chi>" .
        from check_clause[OF ok form'[OF chi] this] 
        have "\<Turnstile> \<chi>" .
      } 
      then have "\<Turnstile> flatten \<psi>" unfolding 1 by (simp add: valid_def)
      also have "\<dots> \<Longleftrightarrow>\<^sub>f \<psi>" by (simp add: form)
      finally show ?thesis by simp
    next
      case Triv
      from trivial_clause_checker[OF clause this] show ?thesis . 
    qed
  }
  then have "\<Turnstile> form_all \<phi>s" by (auto simp: valid_Language)
  also have "... \<Longleftrightarrow>\<^sub>f flatten \<phi>" by (auto simp add: *)
  finally show ?thesis by simp
qed

lemma isOK_simp: "isOK x = (x = Inr ())" by auto

lemma check_formula:
  assumes "isOK(check_formula h \<phi>)" and "formula \<phi>"
  shows "\<Turnstile> \<phi>"
proof-
  have "isOK (check_formula h \<phi>) \<Longrightarrow> formula \<phi> \<Longrightarrow> \<Turnstile> \<phi>"
   and "isOK (check_formula_dist hs pre post \<phi>s) \<Longrightarrow>
        (\<forall>\<phi> \<in> set pre. formula \<phi>) \<Longrightarrow>
        (\<forall>\<phi> \<in> set post. formula \<phi>) \<Longrightarrow>
        (\<forall>\<phi> \<in> set \<phi>s. formula \<phi>) \<Longrightarrow> \<forall> \<phi> \<in> set \<phi>s. \<Turnstile> Disjunction (pre @ \<phi> # post)"
   and "isOK (check_formula_lex_weak hs \<phi> \<phi>s) \<Longrightarrow> formula \<phi> \<Longrightarrow> \<forall>\<phi> \<in> set \<phi>s. formula \<phi> \<Longrightarrow> \<Turnstile> Disjunction (\<phi>#\<phi>s)"
   and "isOK (check_formula_lex hs \<phi> \<phi>s) \<Longrightarrow> formula \<phi> \<Longrightarrow> \<forall>\<phi> \<in> set \<phi>s. formula \<phi> \<Longrightarrow> \<Turnstile> Disjunction (\<phi>#\<phi>s)"
   for h \<phi> hs \<phi>s pre post
  proof (induct h \<phi> and hs pre post \<phi>s and hs \<phi> \<phi>s and hs \<phi> \<phi>s rule:check_formula_check_formula_dist_check_formula_lex_weak_check_formula_lex.induct)
    case (1 \<phi>) with check_valid_formula show ?case by auto
  next
    case (2 h \<phi>)
    then have "formula (simplify \<phi>)" by auto 
    with 2 have "\<Turnstile> simplify \<phi>" by (auto dest: check_clause)
    also have "\<dots> \<Longleftrightarrow>\<^sub>f \<phi>" using 2(2) by simp
    finally show ?case .
  next
    case (3 n h \<phi>)
    show ?case
    proof (cases \<phi>)
      case \<phi>: (Disjunction \<phi>s)
      show ?thesis
      proof (cases "n < length \<phi>s")
        case True
        moreover define pre where "pre = take n \<phi>s"
        moreover define post where "post = drop (Suc n) \<phi>s"
        moreover define \<psi> where "\<psi> = \<phi>s ! n"
        ultimately have split: "\<phi>s = pre @ \<psi> # post"
          by (simp add: id_take_nth_drop)
        from 3(3) have form: "\<forall>\<phi> \<in> set pre. formula \<phi>" "\<forall>\<phi> \<in> set post. formula \<phi>"
          by (auto simp: \<phi> pre_def post_def dest: in_set_takeD in_set_dropD)
        with 3(2)[unfolded \<phi>, simplified, folded pre_def post_def]
        have "\<Turnstile> Disjunction (pre @ False\<^sub>f # post)"
          by (intro 3(1)[OF \<phi> refl _] pre_def post_def, auto simp: check_def)
        then show ?thesis
          by (auto simp: \<phi> split valid_Language)
      next
        case False with 3 \<phi> show ?thesis by auto
      qed
    qed (insert 3, auto simp: check_def)
  next
    case *: (4 n hints \<phi>)
    show ?case
    proof (cases \<phi>)
      case \<phi>: (Disjunction \<phi>s)
      with *(3) have form: "\<forall>\<psi> \<in> set \<phi>s. formula \<psi>" by auto
      show ?thesis
      proof (cases "n < length \<phi>s")
        case [simp]: True
        moreover define pre where "pre = take n \<phi>s"
        moreover define post where "post = drop (Suc n) \<phi>s"
        moreover define \<psi> where "\<psi> = \<phi>s ! n"
        ultimately have split: "\<phi>s = pre @ \<psi> # post"
          by (simp add: id_take_nth_drop)
        from True form have form: "formula \<psi>" "\<forall> \<phi> \<in> set pre. formula \<phi>" "\<forall> \<phi> \<in> set post. formula \<phi>"
          unfolding \<psi>_def pre_def post_def by (auto dest:in_set_takeD in_set_dropD)
        note [symmetric,simp] = pre_def post_def \<psi>_def
        show ?thesis
        proof (cases "\<psi>")
          case [simp]: (Conjunction \<psi>s)
          note \<psi>_def[symmetric,unfolded split this,simp]
          from form have form2: "\<forall>\<rho> \<in> set \<psi>s. formula \<rho>" by auto
          from *(1)[OF \<phi> refl _ pre_def post_def, unfolded check_def, simplified, OF refl] form *(2)
          show ?thesis apply (simp add: \<phi> satisfies_Language) apply (unfold split) apply (auto simp: valid_Language) done
        qed (insert * \<phi>, auto)
      qed (insert * \<phi>, auto)
    qed (insert *, auto)
  next
    case *: (5 h \<psi>)
    from *(2) obtain \<phi> \<phi>s where \<psi>: "\<psi> = Disjunction (\<phi>#\<phi>s)" 
      by (cases \<psi>; cases "disjuncts \<psi>", auto)
    with * show ?case by auto
  next 
    case *: (6 h \<psi>) 
    from *(2) obtain \<phi> \<phi>s where \<psi>: "\<psi> = Disjunction (\<phi>#\<phi>s)" 
      by (cases \<psi>; cases "disjuncts \<psi>", auto)
    with * show ?case by auto
  next
    case (10 \<phi> \<phi>s) then show ?case by (auto dest!: trivial_formula simp: valid_Language)
  next
    case *: (11 hint hints \<phi> \<phi>s)
    have ok: "isOK (check_formula_lex_weak (hint # hints) \<phi> \<phi>s)" by fact
    show ?case
    proof (cases "is_or_and_shape \<phi>")
      case False
      then have "hints = []" "isOK(check_formula hint (Disjunction (\<phi>#\<phi>s)))" using ok by (cases hints, auto)+
      from *(3)[OF False this] show ?thesis using * by auto
    next
      case True
      from True obtain \<phi>1 x where \<phi>: "\<phi> = Disjunction [\<phi>1, x]" 
        by (cases \<phi>; cases "disjuncts \<phi>" rule: list_4_cases, auto)
      from True \<phi> obtain \<phi>2 \<phi>3 where \<phi>: "\<phi> = Disjunction [\<phi>1,Conjunction [\<phi>2,\<phi>3]]" 
        by (cases x; cases "conjuncts x" rule: list_4_cases, auto)
      note IH = *(1-2)[OF True \<phi> refl refl refl refl refl refl refl]
      from ok \<phi> True IH *(5,6) 
      have "\<Turnstile> Disjunction (\<phi>2 # \<phi>s)" and "\<Turnstile> Disjunction (\<phi>3 # \<phi>s)" by auto
      then show ?thesis unfolding \<phi> by (auto simp: valid_Language)
    qed
  next
    case *: (12 hints \<phi> \<phi>s)
    note prems = *(9-)
    show ?case
    proof (cases "is_conj_shape hints \<phi>")
      case True
      then obtain h1 h2 where hints: "hints = [h1,h2]" 
        by (cases hints rule: list_4_cases, auto)      
      from True hints obtain \<phi>1 \<phi>2 where \<phi>: "\<phi> = Conjunction [\<phi>1, \<phi>2]" 
        by (cases \<phi>; cases "conjuncts \<phi>" rule: list_4_cases, auto)
      note IH = *(1-2)[OF True hints refl refl \<phi> refl refl refl]
      from IH prems \<phi> hints show ?thesis by (auto simp:isOK_simp valid_Language) 
    next
      case False
      with prems have disj: "is_disj_shape hints \<phi>" by (auto split: if_splits)
      then obtain h1 hints2 where hints: "hints = h1 # hints2" 
        by (cases hints, auto)      
      from disj obtain x y where \<phi>: "\<phi> = Disjunction [x, y]" 
        by (cases \<phi>; cases "disjuncts \<phi>" rule: list_4_cases, auto)
      from disj \<phi> obtain \<phi>1 \<phi>2 where \<phi>: "\<phi> = Disjunction [Conjunction [\<phi>1,\<phi>2], y]" 
        by (cases x; cases "conjuncts x" rule: list_4_cases, auto)  
      from disj \<phi> obtain \<phi>3 \<phi>4 where \<phi>: "\<phi> = Disjunction [Conjunction [\<phi>1,\<phi>2], Conjunction [\<phi>3,\<phi>4]]" 
        by (cases y; cases "conjuncts y" rule: list_4_cases, auto)  
      note IH = *(3-8)[OF False disj hints \<phi> refl refl refl refl refl refl refl refl refl refl refl]
      note prems = prems[unfolded \<phi> hints]
      show ?thesis
      proof (cases hints2 rule:list_3_cases)
        case Nil
        from IH(1-2)[OF Nil] prems
        show ?thesis by (auto simp: Nil)
      next
        case (1 h2)
        from IH(3-4)[OF 1 refl] prems \<phi> hints
        show ?thesis by (auto simp: 1 isOK_simp valid_Language)
      next
        case (2 h2 h3 hints3)
        from IH(5-6)[OF 2 refl] prems \<phi> hints
        show ?thesis by (auto simp: 2 isOK_simp valid_Language)
      qed
    qed
  qed auto
  from this(1) assms show ?thesis by auto
qed
end

type_synonym ('f,'v,'t) definability_checker = "'v \<Rightarrow> 't \<Rightarrow> ('f,'v,'t) exp formula \<Rightarrow> showsl check"

locale pre_definability_checker =
  prelogic where type_fixer = "TYPE('f\<times>'t\<times>'d)"
  for type_fixer :: "('f\<times>'v\<times>'t\<times>'d) itself"
  and definability_checker :: "('f,'v,'t) definability_checker"

locale definability_checker = pre_definability_checker +
  assumes definability_checker:
    "\<And>x ty \<phi>. isOK (definability_checker x ty \<phi>) \<Longrightarrow> assignment \<alpha> \<Longrightarrow> \<exists>a \<in> Values_of_type ty. \<alpha>((x,ty):=a) \<Turnstile> \<phi>"

lemmas [code] =
  prelogic.literal_def
  pre_trivial_checker.trivial_checker_lit.simps
  pre_trivial_checker.trivial_clause_checker_def
  pre_sorted_algebra.has_type.simps
  pre_logic_checker.check_valid_formula_def
  pre_logic_checker.check_formula.simps
  pre_logic_checker.check_formula_dist.simps
  pre_logic_checker.check_formula_lex_weak.simps
  pre_logic_checker.check_formula_lex.simps
  pre_logic_checker.is_or_and_shape.simps
  pre_logic_checker.is_disj_shape.simps
  pre_logic_checker.is_conj_shape.simps

end
